# Read Vidar .dat file and DVS .aedat file, and merge Vidar spikes and DVS events into an image.
# Jianing Li, Peking University, lijianing@pku.edu.cn.

import argparse
import os
import numpy as np
import xlrd
import math
import cv2
import h5py
from PIL import ImageFont

parser = argparse.ArgumentParser(
    description='Read the .dat file from Vidar and .aedat file from DVS, and then merge them into images.')

parser.add_argument(
    '-n',
    '--input_file',
    default = './statistics/PKU_Vidar_DVS_statistics.xlsx',
    help='the .xlsx file path'
    )

parser.add_argument(
    '-c',
    '--class_label',
    default = './statistics/neuromorphic_object_classes.txt',
    help='the classes for bounding boxes'
    )


def get_sequences_ids(data_path):
    """
    getting ids from .txt file.
    args :
        - data_path: .txt file.
    return:
        - ids: the sequences from .txt file.
    """
    ids = []
    with open(data_path, 'r') as sequence_ids:
        ids.extend(map(str.strip, sequence_ids.readlines()))

    return ids


def get_trans_matrix(trans_matrix_file):
    """
    getting trans matrix.
    args :
        - trans_matrix_file: .txt file.
    return:
        - trans_matrix: numpy, 3*3.
    """
    trans_matrix = []
    with open(trans_matrix_file, 'r') as f:
        data = f.readlines()
        for i, line in enumerate(data):
            tmp = list(map(float, line.split()))
            trans_matrix.append(tmp)

    return np.array(trans_matrix)


def load_raw_spike_data(input_path, width=400, height=250):
    """
    Load raw spike data from .dat file.
    args :
        - spike(0/1).
        - Vidar, width=400, height=250
    return:
        - spike frames: length, width, height
    """
    with open(input_path, 'rb') as f:
        fbytes = f.read()
    fnum = (len(fbytes) * 8) // (width * height)  # number of frames
    spike_frames = np.frombuffer(fbytes, dtype=np.uint8)
    spike_frames = np.array([spike_frames & (1 << i) for i in range(8)])
    spike_frames = spike_frames.astype(np.bool).astype(np.uint8)
    spike_frames = spike_frames.transpose(1, 0).reshape(fnum, height, width)
    spike_frames = np.flip(spike_frames, 1)

    return spike_frames


def window_method(seq, start, length):
    """
    Reconstruction images using constant windows using slow motion scene.
    args :
        - seq: raw spike sequences.
        - start, end are indexes in spike sequences.
    return:
        - images: intensity estimation.
    """
    result = seq[start:start+length].mean(axis=0) * 255
    result = np.power(result,1/2.2)
    result = (255*result/(np.max(result)))

    return result.astype(np.uint8)


def interval_method(seq, mid):
    """
    Reconstruction images using timestamp interval for fast motion scene.
    args :
        - seq: raw spike sequences.
        - mid: reconstruction moment.
    return:
        - images: intensity estimation.
    """
    length, height, width = seq.shape
    result = np.zeros((height, width))
    for i in range(height):
        for j in range(width):
            up, down = mid+1, mid
            for up in range(mid+1, length):
                if seq[up, i, j] == 1:
                    break
            for down in range(mid, -1, -1):
                if seq[down, i, j] == 1:
                    break
            result[i, j] = math.pow(255 / (up - down),1/2.2)
    result = (255*result/(np.max(result)))

    return result.astype(np.uint8)


def recon_ivs(seq, start_moment, end_moment, fps, recon_method):
    """
    Reconstruction images for Vidar camera.
    args :
        - seq: raw spike sequences.
        - start_moment, end_moment: start and end index in spike stream.
        - fps: frames per second.
        - recon_method: providing two intensity estimation methods.
    return:
        - images: intensity estimation.
    """
    recon_interval = int(20000.0/fps) 
    frames = []
    recon_moment = start_moment
    if recon_method == 'itv':
        while(True):
            recon_frame = interval_method(seq, recon_moment)
            #cv2.imwrite('../Vidar_frames/{}.png'.format(recon_moment), recon_frame)
            #print('The {} frame has been done!'.format(recon_moment))
            frames.append(recon_frame)
            recon_moment += recon_interval
            if recon_moment > end_moment:
                break

    elif recon_method == 'win':
        while(True):
            recon_frame = window_method(seq, recon_moment, recon_moment + recon_interval)
            #cv2.imwrite('../Vidar_frames/{}.png'.format(recon_moment), recon_frame)
            #print('The {} frame has been done!'.format(recon_moment))
            frames.append(recon_frame)
            recon_moment += recon_interval
            if recon_moment > end_moment:
                break

    return frames


def dvs_merge_vidar(events, img, width=346, height=260):
    """
    simple display function that shows negative events as blue dots and positive as red one
    on a white background
    args :
        - events structured numpy array: timestamp, x, y, polarity.
        - img (numpy array, height x width x 3) optional array to paint event on.
        - width int.
        - height int.
    return:
        - img numpy array, height x width x 3).
    """

    if events.size:
        assert events['x'].max() < width, "out of bound events: x = {}, w = {}".format(events['x'].max(), width)
        assert events['y'].max() < height, "out of bound events: y = {}, h = {}".format(events['y'].max(), height)

        ON_index = np.where(events['polarity'] == 1)

        img[events['y'][ON_index], events['x'][ON_index], :] = [30, 30, 220] * events['polarity'][ON_index][:, None]  # red

        OFF_index = np.where(events['polarity'] == 0)
        img[events['y'][OFF_index], events['x'][OFF_index], :] = [200, 30, 30] * (events['polarity'][OFF_index] + 1)[:,None]  # blue

    return img


def event_to_bins(input_filename, temporal_length, interval):
    """
    Mapping asynchronous events into event images
    args :
        - input_filename: .hdf5 file, saving dvs events.
        - temporal_length: the total temporal window in DVS event stream.
        - interval: the interval between two frames.
    return:
        - event_bins: split continuous event stream into bins.
    """
    with h5py.File(input_filename) as f:
        events = f['events'][:] # get events

    frame_numbers = temporal_length/interval

    event_bins = []

    for i in range(int(frame_numbers)):
        start_index = np.searchsorted(events['timestamp'], i*interval*1e6) # list B(event_timestamps) insert in list A(timestamp_indexes)
        #end_index = np.searchsorted(events['timestamp'], (i+1)*interval*1e6)
        end_index = start_index + 5000 # 15000 event numbers, rotation
        event_bin = events[start_index:end_index]
        event_bins.append(event_bin)

    return event_bins


def get_bounding_boxes(label_path):
    """
    getting bounding boxes from asynchronous visual streams at 50 Hz.
    args :
        - label_path: the label path of one visual stream.
    return:
        - all_true_boxes: get the labeled bounding boxes at 50 Hz.
    """
    labels_ids = []
    all_true_boxes = []
    with open(label_path, 'r') as sequence_ids:
        labels_ids.extend(map(str.strip, sequence_ids.readlines()))
        for i, sequence_id in enumerate(labels_ids):
            true_bounding_boxes = sequence_id.split(' ')[1:]
            true_bounding_boxes = [(int(box.split(',')[0]), int(box.split(',')[1]), int(box.split(',')[2])-int(box.split(',')[0]), \
                int(box.split(',')[3])-int(box.split(',')[1]),int(box.split(',')[4])) for box in true_bounding_boxes]
            boxes_type={'names':('x', 'y','w','h','class_id'), 'formats':('u4','u4', 'u4','u4','u4')}
            true_bounding_boxes=np.array(true_bounding_boxes,dtype=boxes_type)
            all_true_boxes.append(true_bounding_boxes)

    return all_true_boxes


def draw_true_bboxes(img, boxes, classes):
    """
    drawing the bounding boxes in the merged image.
    args :
        - img: the merged image at a timestamp from Vidar and DVS streams.
        - boxes: the ground truth.
        - classes: all classes in the PKU-Vidar-DVS dataset.
    return:
        - None: an labeled image.
    """
    colors = cv2.applyColorMap(np.arange(0, 255).astype(np.uint8), cv2.COLORMAP_HSV)
    colors = [tuple(*item) for item in colors.tolist()]

    for i in range(boxes.shape[0]):
        pt1 = (int(boxes['x'][i]), int(boxes['y'][i]))
        size = (int(boxes['w'][i]), int(boxes['h'][i]))
        pt2 = (pt1[0] + size[0], pt1[1] + size[1])
        class_id = boxes['class_id'][i]

        color = colors[int(class_id * 165) % 255]  # 60
        class_name = classes[class_id]

        # drawing boxes
        black_color = (10, 10, 10) # black
        font = ImageFont.load_default()
        display_str_width = [font.getsize(ds)[0] for ds in str(class_name)]
        display_str_height = [font.getsize(ds)[1] for ds in str(class_name)]
        total_display_str_width = (sum(display_str_width) + max(display_str_width) * 1.2) * 1.56
        total_display_str_height = max(display_str_height) * 1.6
        cv2.rectangle(img, (int(pt1[0] - 3), int(pt1[1] - total_display_str_height)), (int(pt1[0] + total_display_str_width), int(pt1[1]+3)), color, thickness=-1)
        cv2.rectangle(img, pt1, pt2, color, 2)  # 14
        cv2.putText(img, class_name, (pt1[0] - 1, int(pt1[1] - total_display_str_height / 4)), cv2.FONT_HERSHEY_SIMPLEX, 0.6, black_color, 2)  # 0.5



def _main(args):
    input_file=os.path.expanduser(args.input_file)
    class_filepath = os.path.expanduser(args.class_label)
    neuromorphic_file = xlrd.open_workbook(input_file)
    neuromorphic_information = neuromorphic_file.sheet_by_name('train')
    sequence_names = neuromorphic_information.col_values(0)[1:]
    temporal_lengths = neuromorphic_information.col_values(1)[1:]
    start_indexs = neuromorphic_information.col_values(2)[1:]
    end_indexs = neuromorphic_information.col_values(3)[1:]
    label_classes = get_sequences_ids(class_filepath)

    #for i in range(len(sequence_names)):
    for i in range(0, 1):
        start_index = start_indexs[i]
        end_index = end_indexs[i]
        output_path = './merge_test/{}'.format(sequence_names[i])
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        # read DVS images.
        davis_file = '../train/DVS/{}.hdf5'.format(sequence_names[i])
        interval = 0.01
        event_bins = event_to_bins(davis_file, temporal_lengths[i], interval)

        # read bounding boxes from label files.
        label_file = '../train/labels/{}.txt'.format(sequence_names[i])
        all_bounding_boxes = get_bounding_boxes(label_file)

        # read Vidar images.
        for j in range(int(start_index), int(end_index)+1):
            index = j-int(start_index)
            vidar_file = '../train/Vidar/{}/{}.dat'.format(sequence_names[i],j)
            trans_matrix_filename = './statistics/registration_matrix.npy' # transformation matrix
            trans_matrix = np.load(trans_matrix_filename)
            spike_frames = load_raw_spike_data(vidar_file)
            vidar_frames = recon_ivs(spike_frames, start_moment=0, end_moment=len(spike_frames) - 1, fps=100, recon_method='itv')

            for k in range(len(vidar_frames)):
                trans_frame = cv2.flip(vidar_frames[k], 1) # first trans, then registration

                vidar_reg_frame = cv2.warpPerspective(trans_frame, trans_matrix, (346, 260), flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP)
                vidar_reg_frame = np.expand_dims(vidar_reg_frame, axis=2)
                vidar_reg_frame = np.concatenate((vidar_reg_frame, vidar_reg_frame, vidar_reg_frame), axis=-1)

                merge_image = dvs_merge_vidar(event_bins[2*j+k], vidar_reg_frame, width=346, height=260)            

                # generate merged images with 50 FPS.
                if (2*j+k)%2 == 1:
                    draw_true_bboxes(merge_image, all_bounding_boxes[int((j*2+k+1)/2)-1], label_classes)
                    save_filename = output_path + '/{}.png'.format(int((j*2+k+1)/2))
                    cv2.imwrite(save_filename, merge_image)
                    print('The {} frame has been done!'.format(save_filename))


if __name__ == '__main__':
    _main(parser.parse_args())